# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to create global trend emulations with MESMER.
"""

import os

import joblib
import numpy as np


# TODO: rename because there's actually no emulation involved in this process
# (global trends always come from an external source)
def create_emus_gt(params_gt, preds_gt, cfg, concat_h_f=False, save_emus=True):
    """
    Create global trend (emissions + volcanoes) emulations for specified ensemble type
    and method.

    Parameters
    ----------
    params_gt : dict
        Parameters dictionary

        - ["targ"] (emulated variable, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - ["time"] (1d array of years, np.ndarray)
        - [xx] (additional keys depend on employed method and are listed in
          train_gt_T_method() function)
    preds_gt : dict
        nested dictionary of predictors for global trend with keys

        - [pred][scen]  (1d/2d arrays (time)/(run, time) of predictor for specific scenario)
    cfg : module
        config file containing metadata
    concat_h_f : bool, optional
        determines if historical and future time period is concatenated into a single
        emulation or not, default = False (must be set to false if no historical data
        provided)
    save_emus : bool, optional
        determines if emulation is saved or not, default = True

    Returns
    -------
    emus_gt : dict
        global trend emulations dictionary with keys

        - [scen] (1d array of global trend emulation time series)

    Notes
    -----
    - Assumptions:
        - if no preds_gt needed, pass time as predictor instead to know which scenarios
          to emulate

    """
    # derive necessary scenario names
    pred_names = list(preds_gt.keys())
    scenarios_emus = list(preds_gt[pred_names[0]].keys())

    scens_out_f = list(map(lambda x: x.replace("h-", ""), scenarios_emus))
    # does nothing in case 'h-' not actually included

    if concat_h_f:
        scens_out = scenarios_emus
    else:
        if "h-" in scenarios_emus[0]:
            scens_out = ["hist"] + scens_out_f
        else:
            scens_out = scens_out_f

    # initialize global trend emulations dictionary with scenarios as keys
    emus_gt = {}

    # apply the chosen method
    if "LOWESS" in params_gt["method"]:
        if concat_h_f:
            for scen_out, scen_out_f in zip(scens_out, scens_out_f):
                emus_gt[scen_out] = np.concatenate(
                    [params_gt["hist"], params_gt[scen_out_f]]
                )
        else:
            for scen_out in scens_out:
                emus_gt[scen_out] = params_gt[scen_out]
    else:
        raise ValueError("The chosen method is currently not implemented.")

    # save the global trend emulation if requested
    if save_emus:
        dir_mesmer_emus = cfg.dir_mesmer_emus
        dir_mesmer_emus_gt = dir_mesmer_emus + "global/global_trend/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_emus_gt):
            os.makedirs(dir_mesmer_emus_gt)
            print("created dir:", dir_mesmer_emus_gt)
        filename_parts = [
            "emus_gt",
            params_gt["method"],
            *params_gt["preds"],
            params_gt["targ"],
            params_gt["esm"],
            *scens_out,
        ]
        filename_emus_gt = dir_mesmer_emus_gt + "_".join(filename_parts) + ".pkl"
        joblib.dump(emus_gt, filename_emus_gt)

    return emus_gt



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to train global variability module of MESMER.
"""


import os

import joblib
import numpy as np
from statsmodels.tsa.ar_model import ar_select_order


def train_gv(gv, targ, esm, cfg, save_params=True):
    """
    Derive global variability parameters for a specified method from a specifie
    ensemble type.

    Parameters
    ----------
    gv : dict
        Nested global mean variability dictionary with keys

        - [scen] (2d array (run, time) of globally-averaged variability time series)
    targ : str
        target variable (e.g., "tas")
    esm : str
        associated Earth System Model (e.g., "CanESM2" or "CanESM5")
    cfg : config module
        config file containing metadata
    save_params : bool, optional
        determines if parameters are saved or not, default = True

    Returns
    -------
    params_gv : dict
        dictionary containing the trained parameters for the chosen method / ensemble
        type

        - ["targ"] (emulated variable, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] additional params depend on method employed, specified in
          ``train_gv_T_method()`` function

    Notes
    -----
    - Assumption

        - If historical data is used for training, it has its own scenario.

    """

    # specify necessary variables from config file
    method_gv = cfg.methods[targ]["gv"]
    preds_gv = cfg.preds[targ]["gv"]

    scenarios_tr = list(gv.keys())

    # initialize parameters dictionary and fill in the metadata which does not depend on the applied method
    params_gv = {}
    params_gv["targ"] = targ
    params_gv["esm"] = esm
    params_gv["method"] = method_gv
    params_gv["preds"] = preds_gv
    params_gv["scenarios"] = scenarios_tr

    # apply the chosen method
    if params_gv["method"] == "AR":
        params_gv = train_gv_AR(params_gv, gv)
    else:
        raise ValueError("The chosen method is currently not implemented.")

    # save the global variability paramters if requested
    if save_params:
        dir_mesmer_params = cfg.dir_mesmer_params
        dir_mesmer_params_gv = dir_mesmer_params + "global/global_variability/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_params_gv):
            os.makedirs(dir_mesmer_params_gv)
            print("created dir:", dir_mesmer_params_gv)
        filename_parts = [
            "params_gv",
            method_gv,
            *preds_gv,
            targ,
            esm,
            *scenarios_tr,
        ]
        filename_params_gv = dir_mesmer_params_gv + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_gv, filename_params_gv)

    return params_gv


def train_gv_AR(params_gv, gv):
    """
    Derive AR parameters of global variability under the assumption that gv does not
    depend on the scenario.

    Parameters
    ----------
    params_gv : dict
        parameter dictionary containing keys which do not depend on applied method

        - ["targ"] (variable, i.e., tas or tblend, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, i.e., AR, str)
        - ["scenarios"] (emission scenarios used for training, list of strs)
    gv : dict
        nested global mean temperature variability (volcanic influence removed)
        dictionary with keys

        - [scen] (2d array (nr_runs, nr_ts) of globally-averaged temperature variability
          time series)

    Returns
    -------
    params : dict
        parameter dictionary containing original keys plus

        - ["max_lag"] (maximum lag considered when finding suitable AR model, hardcoded
          to 15 here, int)
        - ["sel_crit"] (selection criterion applied to find suitable AR model, hardcoded
          to Bayesian Information Criterion bic here, str)
        - ["AR_int"] (intercept of the AR model, float)
        - ["AR_coefs"] (coefficients of the AR model for the lags which are contained in
          the selected AR model, list of floats)
        - ["AR_lags"] (AR lags which are contained in the selected AR model, list of
          ints)
        - ["AR_std_innovs"] (standard deviation of the innovations of the selected AR
          model, float)

    Notes
    -----
    - TODO
        - change fct to 1) train AR params on each run individually -> 2) average across
          all runs of specific scen -> 3) all scens
        - learn proper way for bic selection -> eg same process as described above but
          2x: 1x to select order (always take median) 1x to fit params (take mean or
          median again?)

    """

    # put all the global variability time series together in single array
    scenarios = params_gv["scenarios"]  # single entry if ic ensemble, otherwise more

    # assumption: nr_runs per scen and nr_ts for these runs can vary
    nr_samples = 0
    for scen in scenarios:
        nr_runs, nr_ts = gv[scen].shape
        nr_samples += nr_runs * nr_ts

    gv_all = np.zeros(nr_samples)
    i = 0
    for scen in scenarios:
        k = (
            gv[scen].shape[0] * gv[scen].shape[1]
        )  # = nr_runs*nr_ts for this specific scenario
        gv_all[i : i + k] = gv[scen].flatten()
        i += k

    # specifiy parameters employed for AR process fitting
    max_lag = 12  # rather arbitrarily chosen in trade off between allowing enough lags and computational time
    # open to change
    sel_crit = "bic"  # selection criterion

    # select AR order and fit the AR model
    AR_order_sel = ar_select_order(gv_all, maxlag=max_lag, ic=sel_crit)
    AR_model = AR_order_sel.model.fit()

    # fill in the parameter dictionary
    params_gv["max_lag"] = max_lag
    params_gv["sel_crit"] = sel_crit
    params_gv["AR_int"] = AR_model.params[0]  # intercept
    params_gv["AR_coefs"] = AR_model.params[
        1:
    ]  # all coefs which are linked to lags, sorted in same way as ar lags
    params_gv["AR_lags"] = AR_model.ar_lags
    params_gv["AR_std_innovs"] = np.sqrt(
        AR_model.sigma2
    )  # sqrt of variance = standard deviation

    return params_gv



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Collection of miscellaneous utility functions.
"""
# flake8: noqa

from .convert import *
from .select import *



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Collection of functions to create emulations with MESMER.

"""
# flake8: noqa
from .create_emus_gt import *
from .create_emus_gv import *
from .create_emus_lt import *
from .create_emus_lv import *
from .make_realisations import make_realisations
from .merge_emus import *



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to load in mesmer output.
"""


import glob
import os.path
import warnings

import joblib


def load_mesmer_output(
    name,
    cfg,
    method_str="_",
    preds_str="_",
    targs_str="_",
    esm_str="_",
    scen_str="",
    mid_path=None,
):
    """Load saved MESMER output (parameters or emulations).

    Parameters
    ----------
    name : str
        saved MESMER output to load (e.g., "params_lt", "emus_lv", "emus_g")
    cfg : module
        config file containing metadata
    method_str : str, optional
        method (e.g., "OLS")
    preds_str : st, optional
        predictos (e.g., "gttas", "gttas_gttas2")
    targs_str : str, optional
        target variables (e.g., "tas")
    esm_str : str, optional
        Earth System Model (e.g., "CanESM2", "CanESM5")
    scen_str : str, otional
        scenario (e.g., "rcp85", "ssp585", "h-ssp585")
    mid_path : str, optional
        middle part of pathway depending on what exactly want to load (e.g.,
        "local/local_trends")

    Returns
    -------
    dict_out : dict
        loaded MESMER output dictionary

        - [xx] (depending on the loaded output)

    Notes
    -----
    - This function can only load a single .pkl file at every call
    - If multiple files exist for the given input strings, an empty dictionary is
      returned
    - If no file exists for the given input strings, an empty dictionary is returned
    - Also partial strings are accepted (with the exception of esm_str, where the full
      ESM name is needed): e.g.:

        - scen_str="h-" for joint historical + ssp scenarios
        - scen_str="hist" for separated historical + ssp scenarios
    - If no mid_path is provided the default MESMER structure for saved params and emus
      is assumed

    """

    # choose directory depending on whether want to load params or emus
    if "params" in name:
        dir_mesmer = cfg.dir_mesmer_params
    elif "emu" in name:
        dir_mesmer = cfg.dir_mesmer_emus

    # identify middle part of pathway depending on what exactly want to load if not passed as keyword argument
    if mid_path is None:
        if "gt" in name:
            mid_path = os.path.join("global", "global_trend")
        elif "gv" in name:
            mid_path = os.path.join("global", "global_variability")
        elif "g" in name:
            mid_path = "global"
        elif "lt" in name:
            mid_path = os.path.join("local", "local_trends")
        elif "lv" in name:
            mid_path = os.path.join("local", "local_variability")
        elif "l" in name:
            mid_path = "local"

    # to ensure that don't end up with multiple files if multiple ESMs share same beginning of name
    if esm_str != "_":
        esm_str = esm_str + "_"

    path_list = glob.glob(
        os.path.join(
            dir_mesmer,
            mid_path,
            f"{name}*{method_str}*{preds_str}*{targs_str}*{esm_str}*{scen_str}*.pkl",
        )
    )

    # load the requested output dictionary
    if len(path_list) == 1:
        dict_out = joblib.load(path_list[0])
    elif len(path_list) == 0:
        warnings.warn("No such file exists. An empty dictionary will be returned.")
        dict_out = {}
    elif len(path_list) > 1:
        warnings.warn(
            "More than 1 file exists for these critera. "
            "Please be more concrete in your selection."
            " An empty dictionary will be returned.",
        )
        dict_out = {}

    return dict_out



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to train local trends module of MESMER.
"""


import os

import joblib
import numpy as np
from sklearn.linear_model import LinearRegression

from mesmer.calibrate_mesmer.train_utils import train_l_prepare_X_y_wgteq


def train_lt(preds, targs, esm, cfg, save_params=True):
    """
    Derive local trends (i.e., forced response) parameters for given ESM for given set
    of targets and predictors.

    Parameters
    ----------
    preds : dict
        nested dictionary of predictors with keys

        - [pred][scen]  (1d/ 2d arrays (time)/(run, time) of predictor for specific
        scenario)
    targs : dict
        nested dictionary of targets with keys

        - [targ][scen] (3d array (run, time, gp) of target for specific scenario)
    esm : str
        associated Earth System Model (e.g., "CanESM2" or "CanESM5")
    cfg : module
        config file containing metadata
    save_params : bool, optional
        determines if parameters are saved or not, default = True

    Returns
    -------
    params_lt : dict
        dictionary with the trained local trend parameters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["method_each_gp_sep"] (states if method is applied to each grid point
          separately, bool)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] additional params depend on method employed
        - ["full_model_contains_lv"] (whether the full model contains part of the local
          variability module, bool)
    params_lv : dict, optional
        dictionary of local variability paramters which are derived together with the
        local trend parameters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - ["part_model_in_lt"] (states if part of the model is saved in params_lt, bool)
        - ["method_lt_each_gp_sep"] (states if local trends method is applied to each
          grid point separately, bool)

    Notes
    -----
    - Assumptions:
        - all targets use same approach and same predictors
        - each predictor and each target has the same scenarios as keys
        - all available scenarios are used for training
        - in predictor list: local trend predictors belong before local variability
          predictors (if there are any)
        - identified parameters are valid for all training scenarios
        - if historical data is used for training, it has its own scenario
        - either each scenario is given the same weight or each time step
    - TODO:
        - find better way to deal with the assumption that local trend predictors belong
        before local variability predictors (e.g., add check on whether this assumption
        is fulfilled or rewrite code such that no longer necessary)

    """

    targ_names = list(targs.keys())
    targ_name = targ_names[0]  # because same approach for each targ
    pred_names = list(preds.keys())

    # specify necessary variables from config file
    wgt_scen_tr_eq = cfg.wgt_scen_tr_eq
    method_lt = cfg.methods[targ_name]["lt"]
    method_lv = cfg.methods[targ_name]["lv"]
    method_lt_each_gp_sep = cfg.method_lt_each_gp_sep

    preds_lt = []
    preds_lv = []
    # for now only gt/gv implemented, but could easily extend to rt/rv (regional) lt/lv
    # (local)  if wanted such preds
    for pred in pred_names:
        if "gt" in pred:
            preds_lt.append(pred)
        elif "gv" in pred:
            preds_lv.append(pred)

    scenarios_tr = list(targs[targ_name].keys())

    # initialize parameters dictionary and fill in the metadata which does not depend on
    # the applied method
    params_lt = {}
    params_lt["targs"] = targ_names
    params_lt["esm"] = esm
    params_lt["method"] = method_lt
    params_lt["method_each_gp_sep"] = method_lt_each_gp_sep
    params_lt["preds"] = preds_lt
    params_lt["scenarios"] = scenarios_tr

    # check if parameters for local variability module have been derived too, if yes,
    # initialize params_lv
    if method_lt in method_lv:
        params_lt["full_model_contains_lv"] = True

        params_lv = {}
        params_lv["targs"] = targ_names
        params_lv["esm"] = esm
        params_lv["method"] = method_lv
        params_lv["preds"] = preds_lv
        params_lv["scenarios"] = scenarios_tr
        params_lv["part_model_in_lt"] = True
        params_lv["method_lt_each_gp_sep"] = method_lt_each_gp_sep

    else:
        params_lt["full_model_contains_lv"] = False
        params_lv = {}  # only initiate empty dictionary

    # prepare predictors and targets such that they can be ingested into the training
    # function
    X, y, wgt_scen_eq = train_l_prepare_X_y_wgteq(preds, targs)

    # prepare weights for individual runs
    if wgt_scen_tr_eq is False:
        wgt_scen_eq[:] = 1
        # if each scen does not get the same weight, each sample gets it instead

    # train the full model + save it (may also contain lv module parts)
    if method_lt_each_gp_sep and method_lt == "OLS":
        nr_gps = y.shape[1]

        # initialize the regression coefficient dictionaries
        params_lt["intercept"] = {}
        for targ in params_lt["targs"]:
            params_lt["intercept"][targ] = np.zeros(nr_gps)
        for pred in params_lt["preds"]:
            params_lt["coef_" + pred] = {}
            for targ in params_lt["targs"]:
                params_lt["coef_" + pred][targ] = np.zeros(nr_gps)
        if len(params_lv) > 0:
            for pred in params_lv["preds"]:
                params_lv["coef_" + pred] = {}
                for targ in params_lt["targs"]:
                    params_lv["coef_" + pred][targ] = np.zeros(nr_gps)

        # derive the OLS parameters for each gp
        for gp in np.arange(nr_gps):
            reg = LinearRegression().fit(X, y[:, gp, :], wgt_scen_eq)

            for targ_idx, targ in enumerate(params_lt["targs"]):
                params_lt["intercept"][targ][gp] = reg.intercept_[targ_idx]

                coef_idx = 0  # coefficient index
                for pred in params_lt["preds"]:
                    params_lt["coef_" + pred][targ][gp] = reg.coef_[targ_idx, coef_idx]
                    coef_idx += 1

                if len(preds_lv) > 0:  # assumption: coefs of lv are behind coefs of lt
                    for pred in params_lv["preds"]:
                        params_lv["coef_" + pred][targ][gp] = reg.coef_[
                            targ_idx, coef_idx
                        ]
                        coef_idx += 1

    # save the local trend paramters if requested
    if save_params:
        dir_mesmer_params = cfg.dir_mesmer_params
        dir_mesmer_params_lt = dir_mesmer_params + "local/local_trends/"
        # check if folder to save params in exists, if not: make it
        if not os.path.exists(dir_mesmer_params_lt):
            os.makedirs(dir_mesmer_params_lt)
            print("created dir:", dir_mesmer_params_lt)
        filename_parts = [
            "params_lt",
            method_lt,
            *preds_lt,
            *targ_names,
            esm,
            *scenarios_tr,
        ]
        filename_params_lt = dir_mesmer_params_lt + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_lt, filename_params_lt)

        # check if local variability parameters need to be saved too
        # overwrites lv module if already exists, i.e., assumption: lt before lv
        if len(params_lv) > 0:
            dir_mesmer_params_lv = dir_mesmer_params + "local/local_variability/"
            # check if folder to save params in exists, if not: make it
            if not os.path.exists(dir_mesmer_params_lv):
                os.makedirs(dir_mesmer_params_lv)
                print("created dir:", dir_mesmer_params_lv)
        filename_parts = [
            "params_lv",
            method_lv,
            *preds_lv,
            *targ_names,
            esm,
            *scenarios_tr,
        ]
        filename_params_lv = dir_mesmer_params_lv + "_".join(filename_parts) + ".pkl"
        joblib.dump(params_lv, filename_params_lv)

    if len(params_lv) > 0:
        return params_lt, params_lv
    else:
        return params_lt



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/

import numpy as np
import xarray as xr

from .create_emus_gv import create_emus_gv
from .create_emus_lt import create_emus_lt
from .create_emus_lv import create_emus_lv
from .merge_emus import create_emus_l


def make_realisations(
    preds_lt,
    params_lt,
    params_lv,
    params_gv_T,
    time,
    n_realisations,
    seeds,
    land_fractions,
):
    """
    Make climate realisations based on pre-calculated MESMER parameters

    Parameters
    ----------
    preds_lt : dict
        nested dictionary of predictors for local trends with keys

        - [pred][scen] (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)
    params_lt : dict
        dictionary with the trained local trend parameters

        - ["targs"] (emulated variables, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["method_each_gp_sep"] (states if method is applied to each grid point
          separately, bool)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (emission scenarios used for training, list of strs)
        - [xx] (additional params depending on method employed)
        - ["full_model_contains_lv"] (whether the full model contains part of the local
          variability module, bool)
    params_lv : dict
        dictionary with the trained local variability parameters

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method)
    params_gv : dict

        - ["targ"] (variable which is emulated, str)
        - ["esm"] (Earth System Model, str)
        - ["method"] (applied method, str)
        - ["preds"] (predictors, list of strs)
        - ["scenarios"] (scenarios which are used for training, list of strs)
        - [xx] (additional keys depend on employed method and are listed in
          train_gv_T_method() function)
    time : dict

        - ["scenario"] timepoints (1D np.ndarray) used for training of the scenario
          (note that hist and scenario e.g. ssp126 are kept separate)
    n_realisations : int
        Number of realisations to draw
    seeds : dict

        - ["esm"] (dict):
            ["scenario"] (dict):
                ["gv"] (seed for global variability)
                ["lv"] (seed for local variability)
    land_fractions : xr.DataArray
        Land fractions of each cell. Used to convert the MESMER outputs back onto grids.
    """

    class _Config:
        """Workaround to mock the ``cfg`` interface used elsewhere"""

        def __init__(self, n_realisations, seeds):
            self.nr_emus_v = n_realisations
            self.seed = seeds

    cfg = _Config(n_realisations, seeds)

    # TODO: add better checks for what happens if scenarios have different
    # time axis etc.
    a_scenario_key = [k for k in time.keys() if k != "hist"][0]
    time_all = np.concatenate([time["hist"], time[a_scenario_key]])
    esm_gv_T = params_gv_T["esm"]
    time_seeds = seeds[esm_gv_T].keys()
    preds_gv = {"time": {k: time_all for k in time_seeds}}

    emus_gv_T = create_emus_gv(params_gv_T, preds_gv, cfg, save_emus=False)
    preds_lv = {"gvtas": emus_gv_T}

    emus_lt = create_emus_lt(params_lt, preds_lt, cfg, concat_h_f=True, save_emus=False)
    emus_lv = create_emus_lv(params_lv, preds_lv, cfg, save_emus=False)
    emus_l = create_emus_l(emus_lt, emus_lv, params_lt, params_lv, cfg, save_emus=False)

    # xarray DataArrays or DataSets with labelled dimensions
    out = _convert_raw_mesmer_to_xarray(emus_l, land_fractions, time=time)

    return out


def _convert_raw_mesmer_to_xarray(emulations, land_fractions, time):
    land_fractions_stacked = land_fractions.stack(z=("lat", "lon")).dropna("z")

    tmp = []
    for scenario, outputs in emulations.items():
        for variable, values in outputs.items():
            time = np.concatenate([time["hist"], time[scenario.replace("h-", "")]])
            variable_out = (
                land_fractions_stacked.expand_dims({"year": time})
                .expand_dims({"realisation": range(values.shape[0])})
                .copy()
            )
            variable_out.values = values

            variable_out = variable_out.reset_index("z").expand_dims(
                {"scenario": [scenario]}
            )
            variable_out.name = variable

            tmp.append(variable_out)

    out = xr.merge(tmp)

    return out



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
The mesmer package provides tools to train the MESMER emulator, create emulations, and
analyze the results.
"""
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

# flake8: noqa

# import subpackages (so that if I import mesmer, I can directly access e.g., io content
# by calling e.g., mesmer.io.load_const())
from . import calibrate_mesmer, create_emulations, io, utils

# like this could directly access all functions (i.e., if I import mesmer, I can
# directly access fcts, e.g., mesmer.load_const() decided against it for now to keep the
# package structure more visible) from .io import *



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to load in observations which are saved locally.
"""


import numpy as np
import pandas as pd
import xarray as xr


def load_obs(targ, prod, lon, lat, cfg, sel_ref="native", ignore_nans=True):
    """Load observations which you previously downloaded.

    Parameters
    ----------
    targ : str
        target variable (e.g., "tblend")
    prod : str
        product (e.g., "best" or "cw")
    cfg : module
        config file containing metadata
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    sel_ref : str, optional
        selected reference period, (e.g., "native" (original one) or "esm" (the one of
        the esm))
    ignore_nans : bool, optional
        if True global average = average across available gp, if False nan value if not
        all gps available

    Returns
    -------
    var : dict
        variable anomaly dictionary with keys

        - [obs] (4d array (run, time, lat, lon) of variable whereby run=1 because only a
          single realization of real world)
    GVAR : dict
        area-weighted global mean variable anomaly dictionary with keys

        - [obs] (2d array (run, time) of globally-averaged variable anomaly time series)
    time : np.ndarray
        1d array of years

    """

    # define the function mapping
    targ_func_mapping = {"tblend": load_obs_tblend}
    # once start working with other vars, extend dict eg {"pr": load_obs_pr}

    # select the function to apply
    load_targ = targ_func_mapping[targ]

    var, time = load_targ(prod, lon, lat, cfg, sel_ref)

    # compute global average
    lons, lats = np.meshgrid(lon["c"], lat["c"])
    wgt_2d = np.cos(np.deg2rad(lats))
    wgt_3d = np.tile(wgt_2d, [len(time), 1]).reshape(
        [len(time), wgt_2d.shape[0], wgt_2d.shape[1]]
    )
    if (
        ignore_nans
    ):  # grid points with nans are left aside (ie global mean is global mean of non nan gp)
        masked_var = np.ma.masked_array(var, np.isnan(var))
        GVAR = np.ma.average(masked_var, axis=(1, 2), weights=wgt_3d)
    else:  # each field with nans inside will result in nan as global mean
        GVAR = np.average(var, axis=(1, 2), weights=wgt_3d)

    # to adhere to same standards as the ESMs, add "scenario" key ("obs" for observational time period)
    var_dict = {}
    GVAR_dict = {}
    var_dict["obs"] = np.expand_dims(
        var, axis=0
    )  # same format as if were an ESM with a single run
    GVAR_dict["obs"] = np.expand_dims(GVAR, axis=0)

    return var_dict, GVAR_dict, time


def load_obs_tblend(prod, lon, lat, cfg, sel_ref):
    """Load spatially infilled tblend observations. Currently available: best and cw.

    Parameters
    ----------
    prod : str
        product (e.g., "best" or "cw")
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
    cfg : module
        config file containing metadata
    sel_ref : str, optional
        selected reference period, (e.g., "native" (original one) or "esm" (the one of
        the esm))

    Returns
    -------
    var : np.ndarray
        3d array (time, lat, lon) of blended temperatures
    time np.ndarray
        1d array of years

    """

    if prod == "best":
        path = (
            cfg.dir_obs + "blended_temperatures/" + prod + "/best_ann_1850-2019_g025.nc"
        )
        tblend = xr.open_dataset(path).temperature
        tblend["time"] = np.arange(1850, 2020)
    elif prod == "cw":
        path = (
            cfg.dir_obs + "blended_temperatures/" + prod + "/cw_ann_1850-2018_g025.nc"
        )
        tblend = xr.open_dataset(path).temperature_anomaly
        tblend["time"] = np.arange(1850, 2019)

    # check if lon / lat of tblend matches lon / lat of cmip models
    if (tblend.lon != lon["c"]).any() or (tblend.lat != lat["c"]).any():
        raise ValueError(
            "The grids of the ESM output and the observations do not agree."
        )
        tblend = []

    # extract time
    time = tblend.time.values

    # rebaseline and extract array if requested, otherwise just extract array
    if sel_ref == "native":
        tblend = tblend.values
    elif sel_ref == "esm":
        ref = cfg.ref
        tblend = (
            tblend.values
            - tblend.sel(time=slice(ref["start"], ref["end"])).mean(dim="time").values
        )
        # .mean() ignores nan in the selected time slice. Only if all time steps are nans, the mean is a nan too.
    else:
        raise ValueError("No such re-baselining is currently implemented.")

    return tblend, time


def load_strat_aod(time, dir_obs):
    """Load observed global stratospheric aerosol optical depth time series.

    Parameters
    ----------
    time : np.ndarray
        1d array of years the AOD time series is required for
    dir_obs : str
        pathway to observations

    Returns
    -------
    aod_obs : np.ndarray
        1d array of observed global stratospheric AOD time series

    Notes
    -----
    - Assumption: time covers max full extend historical period (i.e., 1850 - 2014 for
      cimp6, 1850 - 2005 for cmip5)
    - potentially TODO: check if want to integrate it into load_obs() fct somehow, but
      likely not as it is quite different from other obs

    """

    path_file = dir_obs + "aerosols/isaod_gl.dat"
    ts = pd.read_csv(
        path_file, delim_whitespace=True, skiprows=11, names=("year", "month", "AOD")
    )
    date_range = pd.date_range(
        *(
            pd.to_datetime(
                [
                    str(ts["year"].iloc[0]) + "-" + str(ts["month"].iloc[0]),
                    str(ts["year"].iloc[-1]) + "-" + str(ts["month"].iloc[-1]),
                ]
            )
            + pd.offsets.MonthEnd()
        ),
        freq="m"
    )
    aod_obs = xr.DataArray(
        ts["AOD"].values, dims=("time",), coords=dict(time=("time", date_range))
    )

    aod_obs = aod_obs.groupby("time.year").mean("time")
    aod_obs = aod_obs.sel(year=slice(str(time[0]), str(time[-1]))).values

    return aod_obs



import versioneer
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

PACKAGE_NAME = "mesmer-emulator"
DESCRIPTION = "Modular Earth System Model Emulator with spatially Resolved output"
KEYWORDS = [
    "climate",
    "atmosphere",
    "Earth System Model Emulator",
]

AUTHOR = "mesmer developpers"
EMAIL = "mesmer@env.ethz.ch"
URL = "https://github.com/MESMER-group/mesmer"  # use documentation url?
PROJECT_URLS = {
    "Source": "https://github.com/MESMER-group/mesmer",
    "Bug Reports": "https://github.com/MESMER-group/mesmer/issues",
    # "Documentation": "TBD",
}

LICENSE = "GPLv3+"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Atmospheric Science",
]


REQUIREMENTS_INSTALL = [
    "dask[complete]",
    "geopy",
    "numpy",
    "pandas",
    "scikit-learn",
    "statsmodels",
    "regionmask",
    "xarray",
]
REQUIREMENTS_TESTS = [
    "pytest",
    "pytest-cov",
    "pytest-xdist",
]
REQUIREMENTS_DEV = [
    "black",
    "flake8",
    "isort",
    "setuptools",
    "twine",
    "wheel",
    *REQUIREMENTS_TESTS,
]

REQUIREMENTS_EXTRAS = {
    "dev": REQUIREMENTS_DEV,
    "tests": REQUIREMENTS_TESTS,
}


SOURCE_DIR = "mesmer"

PACKAGES = find_packages()
PACKAGE_DATA = {}


README = "README.md"

with open(README, "r") as readme_file:
    README_TEXT = readme_file.read()


class Mesmer(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        pytest.main(self.test_args)


cmdclass = versioneer.get_cmdclass()
cmdclass.update({"test": Mesmer})

setup(
    name=PACKAGE_NAME,
    version=versioneer.get_version(),
    description=DESCRIPTION,
    long_description=README_TEXT,
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    project_urls=PROJECT_URLS,
    license=LICENSE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    packages=PACKAGES,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    install_requires=REQUIREMENTS_INSTALL,
    extras_require=REQUIREMENTS_EXTRAS,
    cmdclass=cmdclass,
    # entry_points=ENTRY_POINTS,
)



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to load in constant files such as region information, land-sea mask, area
weights, longitude and latitude information.
"""

import copy as copy
import os

import geopy.distance  # https://geopy.readthedocs.io/en/latest/ # give coord as (lat,lon)
import joblib
import numpy as np
import regionmask as regionmask

from ..utils.regionmaskcompat import mask_percentage
from ..utils.xrcompat import infer_interval_breaks


def gaspari_cohn(r):
    """
    Computes the smooth, exponentially decaying Gaspari-Cohn correlation function for a
    given r.

    Parameters
    ----------
    r : float
        d/L with d = geographical distance in km, L = localisation radius in km

    Returns
    -------
    y : np.ndarray
        Gaspari-Cohn correlation function value for given r

    Notes
    -----
    - Smooth exponentially decaying correlation function which mimics a Gaussian
      distribution but vanishes at r=2, i.e., 2x the localisation radius (L)
    - based on Gaspari-Cohn 1999, QJR (as taken from Carrassi et al 2018, Wiley
      Interdiscip. Rev. Clim. Change)

    """
    r = np.abs(r)

    if r >= 0 and r < 1:
        y = 1 - 5 / 3 * r ** 2 + 5 / 8 * r ** 3 + 1 / 2 * r ** 4 - 1 / 4 * r ** 5
    if r >= 1 and r < 2:
        y = (
            4
            - 5 * r
            + 5 / 3 * r ** 2
            + 5 / 8 * r ** 3
            - 1 / 2 * r ** 4
            + 1 / 12 * r ** 5
            - 2 / (3 * r)
        )
    if r >= 2:
        y = 0

    return y


def load_phi_gc(lon, lat, ls, cfg, L_start=1500, L_end=10000, L_interval=250):
    """
    Loads or creates (if not available yet) distance matrix and Gaspari-Cohn correlation
    matrix.

    Parameters
    ----------
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
        - ["e"] (1d array with longitudes at edges of grid cells)
        - ["grid"] (2d array (lat,lon) of longitudes)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)
        - ["e"] (1d array with latitudes at edges of grid cells)
        - ["grid"] (2d array (lat,lon) of latitudes)
    ls : dict
        land-sea dictionary with keys

        - ["grid_raw"] (2d array (lat,lon) of subsampled land fraction)
        - ["grid_no_ANT"] (grid_raw with Antarctica removed)
        - ["gp_l"] (1d array of fraction of land at land grid points)
        - ["grid_l"] (2d array (lat,lon) of fraction of land at land grid points)
        - ["idx_grid_l"] (2d boolean array (lat,lon) with land grid points = True for
          plotting on map)
        - ["grid_l_m"] (2d masked array (lat,lon) with ocean masked out for plotting on
          map)
        - ["wgt_gp_l"] (1d array of land area weights, i.e.,
          area weight * land fraction)
    cfg : module
        config file containing metadata
    L_start : int, optional
        smallest localisation radius which is tested
    L_end : int, optional
        largest localisation radius which is tested
    L_interval : int, optional
        spacing interval between tested localisation radii

    Returns
    -------
    phi_gc : np.ndarray
        2d array (gp, gp) of Gaspari-Cohn correlation matrix for grid points used for
        covariance localisation

    Notes
    -----
    - If no complete number of L_intervals fits between L_start and L_end, L_intervals
      are repeated until the closest possible L value below L_end is reached.
    - L_end should not exceed 10000 by much because eventually ValueError: the input matrix must be positive semidefinite in train_lv())
    """

    dir_aux = cfg.dir_aux
    threshold_land = cfg.threshold_land

    L_set = np.arange(L_start, L_end + 1, L_interval)

    # geodistance for all gps for certain threshold
    geodist_name = "geodist_landthres_{tl:1.2f}.pkl".format(tl=threshold_land)
    if not os.path.exists(dir_aux + geodist_name):
        # create geodist matrix + save it
        print("compute geographical distance between all land points")
        nr_gp_l = ls["idx_grid_l"].sum()
        lon_l_vec = lon["grid"][ls["idx_grid_l"]]
        lat_l_vec = lat["grid"][ls["idx_grid_l"]]
        geodist = np.zeros([nr_gp_l, nr_gp_l])

        # could be sped up by only computing upper or lower half of matrix
        # since only needs to be done 1x for every land threshold, not implemented
        for i in np.arange(nr_gp_l):
            for j in np.arange(nr_gp_l):
                geodist[i, j] = geopy.distance.distance(
                    (lat_l_vec[i], lon_l_vec[i]), (lat_l_vec[j], lon_l_vec[j])
                ).km

            if i % 200 == 0:
                print("done with gp", i)

        # create auxiliary directory if does not exist already (else leave directory unaltered)
        os.makedirs(dir_aux, exist_ok=True)

        # save the geodist file
        joblib.dump(geodist, dir_aux + geodist_name)

    else:
        # load geodist matrix
        geodist = joblib.load(dir_aux + geodist_name)

    # gaspari-cohn correlation function phi
    phi_gc_name = "phi_gaspari-cohn_landthres_{tl:1.2f}_Lset_{L_start}-{L_interval}-{L_end}.pkl".format(
        tl=threshold_land, L_start=L_start, L_interval=L_interval, L_end=L_end
    )

    if not os.path.exists(dir_aux + phi_gc_name):
        print(
            "compute Gaspari-Cohn correlation function phi for a number of localization radii"
        )

        phi_gc = {}
        for L in L_set:
            phi_gc[L] = np.zeros(geodist.shape)
            for i in np.arange(geodist.shape[0]):
                for j in np.arange(geodist.shape[1]):
                    phi_gc[L][i, j] = gaspari_cohn(geodist[i, j] / L)
            print("done with L:", L)

        joblib.dump(phi_gc, dir_aux + phi_gc_name)

    else:
        phi_gc = joblib.load(dir_aux + phi_gc_name)

    return phi_gc


def load_regs_ls_wgt_lon_lat(reg_type, lon, lat):
    """Load constant files.

    Parameters
    ----------
    reg_type : str
        region type ("ar6.land", "countries", "srex")
    lon : dict
        longitude dictionary with key

        - ["c"] (1d array with longitudes at center of grid cell)
    lat : dict
        latitude dictionary with key

        - ["c"] (1d array with latitudes at center of grid cell)

    Returns
    -------
    reg_dict : dict
        region dictionary with keys

        - ["type"] (region type)
        - ["abbrevs"] (abbreviations for regions)
        - ["names"] (full names of regions)
        - ["grids"] (3d array (region, lat, lon) of subsampled region fraction)
        - ["grid_b"] (2d array (lat, lon) of regions with each grid point being assigned
          to a single region ("binary" grid))
        - ["full"] (full Region object (for plotting region outlines))
    ls : dict
        land-sea dictionary with keys

        - ["grid_raw"] (2d array (lat, lon) of subsampled land fraction)
        - ["grid_no_ANT"] (grid_raw with Antarctica removed)
    wgt : np.ndarray
        2d array (lat,lon) of weights to be used for area weighted means
    lon : dict
        longitude dictionary with added keys

        - ["e"] (1d array with longitudes at edges of grid cells)
        - ["grid"] (2d array (lat,lon) of longitudes)
    lat : dict
        latitude dictionary with added keys

        - ["e"] (1d array with latitudes at edges of grid cells)
        - ["grid"] (2d array (lat,lon) of latitudes)

    Notes
    -----
    - If additional region types are added in this function,
      mesmer.utils.select.extract_land() needs to be adapted too

    """

    # choose the Regions object depending on the region type
    if reg_type == "countries":
        reg = regionmask.defined_regions.natural_earth.countries_110
    elif reg_type == "srex":
        reg = regionmask.defined_regions.srex
    elif reg_type == "ar6.land":
        reg = regionmask.defined_regions.ar6.land

    # extract all the desired information from the Regions object
    reg_dict = {}
    reg_dict["type"] = reg_type
    reg_dict["abbrevs"] = reg.abbrevs
    reg_dict["names"] = reg.names
    reg_dict["grids"] = mask_percentage(
        reg, lon["c"], lat["c"]
    ).values  # have fraction of grid cells
    reg_dict["grid_b"] = reg.mask(
        lon["c"], lat["c"]
    ).values  # not sure yet if needed: "binary" grid with each grid point assigned to single country
    reg_dict[
        "full"
    ] = reg  # to be used for plotting outlines (mainly useful for srex regs)

    # obtain a (subsampled) land-sea mask
    ls = {}
    ls["grid_raw"] = np.squeeze(
        mask_percentage(
            regionmask.defined_regions.natural_earth.land_110, lon["c"], lat["c"]
        ).values
    )
    # gives fraction of land -> in extract_land() script decide above which land fraction threshold to consider a grid point as a land grid point

    # remove Antarctica
    idx_ANT = np.where(lat["c"] < -60)[0]
    ls["grid_no_ANT"] = copy.deepcopy(ls["grid_raw"])
    ls["grid_no_ANT"][idx_ANT] = 0  #

    # derive the weights
    lon["grid"], lat["grid"] = np.meshgrid(lon["c"], lat["c"])
    wgt = np.cos(np.deg2rad(lat["grid"]))

    # derive longitude / latitude of edges of grid cells for plotting with pcolormesh
    lon["e"], lat["e"] = infer_interval_breaks(lon["c"], lat["c"])

    return reg_dict, ls, wgt, lon, lat



# MESMER, land-climate dynamics group, S.I. Seneviratne
# Copyright (c) 2021 ETH Zurich, MESMER contributors listed in AUTHORS.
# Licensed under the GNU General Public License v3.0 or later see LICENSE or
# https://www.gnu.org/licenses/
"""
Functions to aid the training of MESMER.
"""


import numpy as np


def train_l_prepare_X_y_wgteq(preds, targs):
    """
    Create single array of predictors, single array of targets, and single array of
    weights.

    Parameters
    ----------
    preds : dict
        empty dictionary if none, else nested dictionary of predictors with keys

        - [pred][scen]  (1d/ 2d arrays (time)/(run, time) of predictor for specific scenario)
    targs : dict
        nested dictionary of targets with keys

        - [targ][scen] (3d array (run, time, gp) of target for specific scenario)

    Returns
    -------
    X : np.ndarray
        empty array if none, else 2d array (sample, pred) of predictors
    y : np.ndarray
        3d array (sample, gp, targ) of targets
    wgt_scen_eq : np.ndarray
        1d array (sample) of sample weights based on equal treatment of each scenario
        (if scen has more samples, each sample gets less weight)
    """

    targ_names = list(targs.keys())
    targ_name = targ_names[0]  # because same approach for each targ
    pred_names = list(preds.keys())

    # identify characteristics of the predictors and the targets
    targ = targs[
        targ_name
    ]  # predictors are not influenced by whether there is a single or there are multiple targets
    scens = list(targ.keys())

    # assumption: nr_runs per scen and nr_ts for these runs can vary
    # derive weights such that each scenario receives same weight (divide by nr samples)
    nr_samples = 0
    wgt_scen_eq = []
    for scen in scens:
        nr_runs, nr_ts, nr_gps = targ[scen].shape
        nr_samples_scen = nr_runs * nr_ts
        wgt_scen_eq = np.append(wgt_scen_eq, np.repeat(1 / nr_runs, nr_samples_scen))
        nr_samples += nr_samples_scen

    nr_preds = len(pred_names)
    nr_targs = len(targ_names)

    # derive X (ie array of predictors)
    if nr_preds == 0:
        X = np.empty(0)
    else:
        X = np.zeros([nr_samples, nr_preds])
        for p in np.arange(nr_preds):  # index for predictors
            pred_name = pred_names[p]  # name of predictor p
            s = 0  # index for samples
            pred_raw = preds[pred_name]  # values of predictor p
            for scen in scens:
                if (
                    len(pred_raw[scen].shape) == 2
                ):  # if 1 time series per run for predictor (e.g., gv)
                    k = (
                        pred_raw[scen].shape[0] * pred_raw[scen].shape[1]
                    )  # nr_runs*nr_ts for this specific scenario
                    X[s : s + k, p] = pred_raw[scen].flatten()
                    s += k
                elif (
                    len(pred_raw[scen].shape) == 1
                ):  # if single time series as predictor (e.g. gt): repeat ts as many times as runs available
                    nr_runs, nr_ts, nr_gps = targ[scen].shape
                    nr_samples_scen = nr_runs * nr_ts
                    X[s : s + nr_samples_scen, p] = np.tile(pred_raw[scen], nr_runs)
                    s += nr_samples_scen
                else:
                    raise ValueError("Predictors of this shape cannot be processed.")

    # derive y (ie array of targets)
    y = np.zeros([nr_samples, nr_gps, nr_targs])
    for t, targ_name in enumerate(targ_names):
        targ = targs[targ_name]
        s = 0
        for scen in scens:
            k = (
                targ[scen].shape[0] * targ[scen].shape[1]
            )  # nr_runs*nr_ts for this scenario
            y[s : s + k, :, t] = targ[scen].reshape(k, -1)
            s += k

    return X, y, wgt_scen_eq
